//  AuthenticatorKeyPairGenerator.java
//
//  Last modified 2/16/99 by Alan Frindell
//
//  This class generates the public/private key pair used in the
// authentication protocol you are implementing.
//  
//  It is OK to use the default parameters.

package Authenticator;

import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;

public class AuthenticatorKeyPairGenerator extends KeyPairGenerator {

  private int strength;
  private SecureRandom source;
  
  /** p and q are prime with pr = (1 - 1 / 2^CONFIDENCE) */
  private static final int CONFIDENCE = 80;

  /** Default number of bits for the modulus */
  private static final int DEFAULT_STRENGTH = 1024;
  
  
  public AuthenticatorKeyPairGenerator() { 
    super("Authenticator"); 
  }
  
  public void initialize () {
    initialize(DEFAULT_STRENGTH, null);
  }
  
  public void initialize (int strength, SecureRandom source) {
    this.strength = (strength < 2) ? DEFAULT_STRENGTH : strength;
    try {
      this.source = (source == null) ? SecureRandom.getInstance("SHA1PRNG"):
	source;
    } catch (Exception e) {
      System.out.println("KeyPairGenerator: " + e.getMessage());
      e.printStackTrace();
    }
      
  }
  
  public KeyPair generateKeyPair() {
    int k1 = strength / 2;
    int k2 = strength - k1;
    BigInteger p, q, N, s, R, T, t, phi, x;

    try {
      //  A prime, so claims Dan.
      s = new BigInteger("1000003");
      
      do {
	p = getHappyPrime(k1, s, source);
	System.out.println("Got p");
	q = getHappyPrime(k2, s, source);
	System.out.println("Got q");
      } while (p.equals(q));
      
      N = p.multiply(q);
      phi = (p.subtract(BigInteger.ONE)).multiply(q.subtract(BigInteger.ONE));

      x = (new BigInteger(strength, source)).mod(phi);

      R = x.modPow(phi.divide(s.pow(2)), N);

      //  r^s = 1 (mod N) st s != 1 (mod p) and s != 1 (mod q)
      if (!R.modPow(s, N).equals(BigInteger.ONE) || 
      	  R.mod(p).equals(BigInteger.ONE) || 
	  R.mod(q).equals(BigInteger.ONE)) {
      	System.out.println("Doh!");
      } else {
      	System.out.println("Yay");
      }

      t = (new BigInteger(strength, source)).mod(N);
      
      T = t.modPow(s, N);
      
      return new KeyPair(new AuthenticatorPublicKey(N, T, s), 
			 new AuthenticatorPrivateKey(R, p, t));
    } catch (Exception e) {
      System.out.println("Error generating Authenticator params: " + 
			 e.getMessage());
      e.printStackTrace();
    }
    return null;
  }
  
  public BigInteger getHappyPrime(int strength, BigInteger s, 
				  SecureRandom rng) {
    BigInteger j, p, k;
    
    while (true) {
      j = new BigInteger(strength, rng);
      p = j.multiply(s).add(BigInteger.ONE);
      System.out.print("**Searching**");
      while (true) {
	p = p.add(s);
	System.out.print(".");
	k = p.subtract(BigInteger.ONE);
	if (!(k.mod(s.pow(2))).equals(BigInteger.ZERO) && 
	    p.isProbablePrime(CONFIDENCE)) {
	  return p;
	}
	if (j.bitCount() > strength + 2) {
	  break;
	}
      }
    }
  }

}
